@@ -19,7 +19,9 @@
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 #
 from papyon.gnet.constants import *
+from papyon.gnet.errors import *
 from papyon.gnet.resolver import *
+from papyon.util.async import run
 from abstract import AbstractClient
 
 import gobject
@@ -30,12 +32,12 @@ __all__ = ['GIOChannelClient']
 
 class OutgoingPacket(object):
     """Represents a packet to be sent over the IO channel"""
-    def __init__(self, buffer, size, callback=None, *cb_args):
+    def __init__(self, buffer, size, callback=None, errback=None):
         self.buffer = buffer
         self.size = size
         self._sent = 0
         self._callback = callback
-        self._callback_args = cb_args
+        self._errback = errback
 
     def read(self, size=2048):
         if size is not None:
@@ -52,8 +54,11 @@ class OutgoingPacket(object):
 
     def callback(self):
         """Run the callback function if supplied"""
-        if self._callback is not None:
-            self._callback(*self._callback_args)
+        run(self._callback)
+
+    def errback(self, error):
+        """Run the errback function if supplied"""
+        run(self._errback, error)
 
 
 class GIOChannelClient(AbstractClient):
@@ -88,15 +93,10 @@ class GIOChannelClient(AbstractClient):
 
     def _open(self, host, port):
         resolver = HostnameResolver()
-        resolver.query(host, (self.__open, host, port))
+        resolver.query(host, (self.__open, host, port), (self.__open_failed,))
 
     def __open(self, resolve_response, host, port):
-        if resolve_response.status != 0:
-            self.emit("error", IoError.CONNECTION_FAILED)
-            self._transport.close()
-            return
-        else:
-            host = resolve_response.answer[0][1]
+        host = resolve_response.answer[0][1]
 
         # Even though connect_ex *shouldn't* raise an exception,
         # sometimes it does, which is just great.
@@ -112,9 +112,13 @@ class GIOChannelClient(AbstractClient):
             return
         elif err in (EHOSTUNREACH, EHOSTDOWN, ECONNREFUSED, ECONNABORTED,
                 ENETUNREACH, ENETDOWN, EBADFD):
-            self.emit("error", IoError.CONNECTION_FAILED)
+            self.emit("error", IoConnectionFailed(self, str(err)))
             self._transport.close()
 
+    def __open_failed(self, error):
+        self.emit("error", error)
+        self._transport.close()
+
     # convenience methods
     def _watch_remove(self):
         if self._source_id is not None:
@@ -152,6 +156,11 @@ class GIOChannelClient(AbstractClient):
         if self._status in (IoStatus.CLOSING, IoStatus.CLOSED):
             return
         self._status = IoStatus.CLOSING
+
+        for packet in self._outgoing_queue:
+            packet.errback(IoConnectionClosed(self, self._status))
+        self._outgoing_queue = []
+
         self._watch_remove()
         try:
             self._channel.close()
@@ -168,9 +177,11 @@ class GIOChannelClient(AbstractClient):
             return
         self._watch_remove()
 
-    def send(self, buffer, callback=None, *args):
-        assert(self._status == IoStatus.OPEN), self._status
+    def send(self, buffer, callback=None, errback=None):
+        if self._status != IoStatus.OPEN:
+            run(errback, IoConnectionClosed(self, self._status))
+            return
         self._outgoing_queue.append(OutgoingPacket(buffer, len(buffer),
-            callback, *args))
+            callback, errback))
         self._watch_add_cond(gobject.IO_OUT)
 gobject.type_register(GIOChannelClient)
